//
// Parallel Computing, DCC/FCUP
//
// Author: Eduardo R. B. Marques
//
// Test program for bstack_t.
//
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include "bstack.h"

typedef struct {
  bstack_t stack;
  int n_producers;
  int n_consumers; 
  int items;
} shared_data_t;

shared_data_t sd;

void* consumer(void* arg) {
  long rank = (long) arg;
  int count = sd.items / sd.n_consumers;
  for (int i=0; i < count; i++) {
    printf("%ld - consuming %d - %d to go\n", rank, i, count - i);
    long v = (long) bstack_remove(&sd.stack);
    printf("%ld - consumed %ld\n", rank, v);
  }
  printf("%ld - done\n", rank);
  return 0;
}
void* producer(void* arg) {
  long rank = (long) arg;
  int count = sd.items / sd.n_producers;
  for (int i=0; i < count; i++) {
    long v = count * rank + i;
    printf("%ld - producing %d - %d to go\n", rank, i, count - i);
    bstack_add(&sd.stack, (void*) v);
    printf("%ld - produced %ld\n", rank, v);
  }
  printf("%ld - done\n", rank);
  return 0;
}

int main(int argc, char** argv) {
  --argc; ++argv;
  if (argc != 2) {
    printf("Usage: bstack_test.bin <producers> <consumers>\n");
    return 1;
  }
  bstack_init(&sd.stack);
  sd.n_producers = atoi(argv[0]);
  sd.n_consumers = atoi(argv[1]);
  sd.items = 3 * sd.n_producers * sd.n_consumers;
  int total_threads = sd.n_consumers + sd.n_producers;
  pthread_t* vth = (pthread_t*) malloc(sizeof(pthread_t) * total_threads);
  for (long rank = 0; rank < total_threads; rank++) {
    pthread_create(&vth[rank], NULL, 
                   rank < sd.n_producers ? &producer : &consumer, 
                   (void*) rank);
  }
  for (long rank = 0; rank < total_threads; rank++) {
    pthread_join(vth[rank], NULL);
  }
  free(vth);
  bstack_destroy(&sd.stack);
  return 0; 
}
